//---------------------------------------------------------------------------
#pragma hdrstop

#include "uTank.h"
#include "IdGlobal.hpp"

//---------------------------------------------------------------------------
#pragma package(smart_init)
//--------------------------------------------------------------------------
void fnRGB2HSV(int iR, int iG, int iB, float &fH, float &fS, float &fV);
void fnRGB2HSL(int iR, int iG, int iB, float &fH, float &fS, float &fL);
void fnHSV2RGB(float fH, float fS, float fV, int &iR, int &iG, int &iB);
void fnHSL2RGB(float fH, float fS, float fL, int &iR, int &iG, int &iB);
TDateTime fnGetNow();
//---------------------------------------------------------------------------
TDateTime fnGetNow()
{
   TDateTime Tms = EncodeTime(0, 0, 0, 1);//millisecond
   double dConsume=0;
   LARGE_INTEGER iFreq={0};
   LARGE_INTEGER iNow={0};
	QueryPerformanceFrequency(&iFreq);
   QueryPerformanceCounter(&iNow);
   dConsume = (iNow.QuadPart * 1000.)/iFreq.QuadPart;
   Tms = dConsume*Tms.Val;
   return Tms;
}
//----------------------------------------------------------------------------
void fnRGB2HSL(int iR, int iG, int iB, float &fH, float &fS, float &fL)
{
   float fGap = 0.001;
   float fMin, fMax, fR, fG, fB;
   fR = iR/(255.);
   fG = iG/(255.);
   fB = iB/(255.);
   fMax = Max(Max(iR, iG), iB)/255.;
   fMin = Min(Min(iR, iG), iB)/255.;
   if(fabs(fMax-fMin) <= fGap)
      fH = 0;
   else if(fabs(fMax-fR) <= fGap){
      if(fG >= fB)
         fH = 60.*(fG-fB)/(fMax-fMin);
      else
         fH = 360.+ 60.*(fG-fB)/(fMax-fMin);
   }
   else if(fabs(fMax-fG) <= fGap)
      fH =120.+ 60.*(fB-fR)/(fMax-fMin);
   else if(fabs(fMax-fB) <= fGap)
      fH =240.+ 60.*(fR-fG)/(fMax-fMin);
   else
      fH = 0;
   fL = (fMax+fMin)*0.5;

   if((fL == 0) || (fMax==fMin))
      fS = 0;
   else if(fL >0 && fL <= 0.5)
      fS = (fMax-fMin)/(fMax+fMin);
   else
      fS = (fMax-fMin)/(2.-(fMax+fMin));
}
//---------------------------------------------------------------------------
void fnHSL2RGB(float fH, float fS, float fL, int &iR, int &iG, int &iB)
{
   float fR,fG,fB;
   float fTemp1, fTemp2, fTempR, fTempG, fTempB;
   float fTempH, fTempS, fTempL;
   fTempH = fH / 360.0;  //Hue is represented as a range of 360 degrees
   fTempS = fS ;  //Saturation
   fTempL = fL ;  //Lightness

   if(fTempS == 0)           //Saturation of 0 means a shade of grey
    	fR = fG = fB = fTempL;
   else{
      if(fTempL < 0.5)
         fTemp2 = fTempL * (1.0 + fTempS);
      else
         fTemp2 = (fTempL + fTempS) - (fTempL * fTempS);

      fTemp1 = 2.0 * fTempL - fTemp2;
      fTempR = fTempH + 1.0 / 3.0;
      if(fTempR > 1.0)
         fTempR -= 1.0;
      fTempG = fTempH;
      fTempB = fTempH - 1.0 / 3.0;
      if(fTempB < 0)
         fTempB += 1.0;

     // Calculate red value:
      if(6.0 * fTempR < 1.0)
         fR = fTemp1 + (fTemp2 - fTemp1) * 6.0 * fTempR;
      else if(2.0 * fTempR < 1.0)
         fR = fTemp2;
      else if (3.0 * fTempR < 2.0)
         fR = fTemp1 + (fTemp2 - fTemp1) * ((2.0 / 3.0) - fTempR) * 6.0;
      else
         fR = fTemp1;
     // Calculate green value
      if (6.0 * fTempG < 1.0)
         fG = fTemp1 + (fTemp2 - fTemp1) * 6.0 * fTempG;
      else if(2.0 * fTempG < 1.0)
         fG = fTemp2;
      else if(3.0 * fTempG < 2.0)
         fG = fTemp1 + (fTemp2 - fTemp1) * ((2.0 / 3.0) - fTempG) * 6.0;
      else
         fG = fTemp1;
     // Calculate blue value
      if(6.0 * fTempB < 1.0)
         fB = fTemp1 + (fTemp2 - fTemp1) * 6.0 * fTempB;
      else if(2.0 * fTempB < 1.0)
         fB = fTemp2;
      else if(3.0 * fTempB < 2.0)
         fB = fTemp1 + (fTemp2 - fTemp1) * ((2.0 / 3.0) - fTempB) * 6.0;
      else
         fB = fTemp1;
    }
    iR = fR*255.;
    iG = fG*255.;
    iB = fB*255.;
}
//---------------------------------------------------------------------------------------
void fnHSV2RGB(float fH, float fS, float fV, int &iR, int &iG, int &iB)
{
   int iHi ;
   float fF, fP, fQ, fT;
   iHi = floor(fH/60.);
   fF = (fH/60.) - iHi;
   fP = fV * (1.0 - fS);
   fQ = fV * (1.0 - fF*fS);
   fT = fV * (1.0 - (1.0 - fF)*fS);
   switch (iHi){
      case 0: iR = fV*255.; iG = fT*255.; iB = fP*255.; break;
      case 1: iR = fQ*255.; iG = fV*255.; iB = fP*255.; break;
      case 2: iR = fP*255.; iG = fV*255.; iB = fT*255.; break;
      case 3: iR = fP*255.; iG = fQ*255.; iB = fV*255.; break;
      case 4: iR = fT*255.; iG = fP*255.; iB = fV*255.; break;
      case 5: iR = fV*255.; iG = fP*255.; iB = fQ*255.; break;
   }
}
//-------------------------------------------------------------------------------------------
void fnRGB2HSV(int iR, int iG, int iB, float &fH, float &fS, float &fV)
{
   float fGap = 0.001;
   float fMin, fMax, fR, fG, fB;
   fR = iR/(255.);
   fG = iG/(255.);
   fB = iB/(255.);

   fMax = Max(Max(iR, iG), iB)/255.;
   fMin = Min(Min(iR, iG), iB)/255.;
   if(fabs(fMax-fMin) <= fGap)
      fH = 0;
   else if(fabs(fMax-fR) <= fGap){
      if(fG >= fB)
         fH = 60.*(fG-fB)/(fMax-fMin);
      else
         fH = 360.+ 60.*(fG-fB)/(fMax-fMin);
   }
   else if(fabs(fMax-fG) <= fGap)
      fH =120.+ 60.*(fB-fR)/(fMax-fMin);
   else if(fabs(fMax-fB) <= fGap)
      fH =240.+ 60.*(fR-fG)/(fMax-fMin);
   else
      fH = 0;
   if(fMax == 0)
      fS = 0;
   else
      fS = 1.0 - (fMin/fMax);
  fV = fMax;
}
//-------------------------------------------------------------------------
_ImgMatrix::_ImgMatrix()
{
   m_iSysCx = GetSystemMetrics(SM_CXSCREEN);
   m_iSysCy = GetSystemMetrics(SM_CYSCREEN);
   m_pppBmp=0;
   m_pppJpg=0;
   m_pppBmpMem=0;
   m_pppJpgMem=0;
   m_iTop=0;
   m_iLeft=0;
   m_iW=0;
   m_iH=0;
   m_iCmpQual=100;
}
//--------------------------------------------------------------------------
_ImgMatrix::~_ImgMatrix()
{
   fnClear();
}
//------------------------------------------------------------------------
void _ImgMatrix::fnSetJpgQul(int iVar)
{
   m_iCmpQual = iVar;
}
//---------------------------------------------------------------------------
void _ImgMatrix::fnCreate()
{
   fnCreateMem(m_pppBmpMem);
   fnCreateMem(m_pppJpgMem);
   fnCreateBmp(m_pppBmp);
   fnCreateJpg(m_pppJpg);
}
//-------------------------------------
void _ImgMatrix::fnClear()
{
   fnClearMem(m_pppBmpMem);
   fnClearMem(m_pppJpgMem);
   fnClearBmp(m_pppBmp);
   fnClearJpg(m_pppJpg);
}
//-------------------------------------------------------------------
void _ImgMatrix::fnInitMtr(int iColLen, int iRowLen)
{
   m_iColLen = iColLen;
   m_iRowLen = iRowLen;
   fnClear();
   fnCreate();
   m_iColLenPre = m_iColLen;
   m_iRowLenPre = m_iRowLen;
}
//----------------------------------------------------------------------------
void _ImgMatrix::fnGetMtrInf(int &iColLen, int &iRowLen, int &iW, int &iH)
{
   iColLen = m_iColLen;
   iRowLen = m_iRowLen;
   iW = m_iW;
   iH = m_iH;
}
//-------------------------------------------------------------------------
void _ImgMatrix::fnCreateMem(TMemoryStream ***&ppp)
{
   ppp = new TMemoryStream** [m_iColLen];
   for (int i=0;i<m_iColLen ; i++)
      ppp[i] = new TMemoryStream* [m_iRowLen];
   for (int i=0;i<m_iColLen ;i++){
      for (int j=0;j<m_iRowLen ;j++)
         ppp[i][j] = new TMemoryStream;
   }
}
//------------------------------------------------------------------------
void _ImgMatrix::fnCreateBmp(TImage ***&ppp)
{
   ppp = new TImage** [m_iColLen];
   for (int i=0;i<m_iColLen ;i++)
      ppp[i] = new TImage* [m_iRowLen];
   for (int i=0;i<m_iColLen ;i++){
      for (int j=0;j<m_iRowLen ;j++)
         ppp[i][j] = new TImage(0);
   }
}
//--------------------------------------------------
void _ImgMatrix::fnCreateJpg(TJPEGImage ***&ppp)
{
   ppp = new TJPEGImage** [m_iColLen];
   for (int i=0;i<m_iColLen ;i++)
      ppp[i] = new TJPEGImage* [m_iRowLen];
   for (int i=0;i<m_iColLen ;i++){
      for (int j=0;j<m_iRowLen ;j++)
         ppp[i][j] = new TJPEGImage();
   }
}
//-------------------------------------------------------
void _ImgMatrix::fnClearMem(TMemoryStream ***&ppp)
{
   if(!ppp) return;
   for (int i=0;i<m_iColLenPre ;i++){
      for (int j=0;j<m_iRowLenPre ;j++)
        delete ppp[i][j];
   }
   for (int i=0;i<m_iColLenPre ;i++){
      delete [] ppp[i];
   }
   delete [] ppp;
   ppp = 0;
}
//----------------------------------------------------------------
void _ImgMatrix::fnClearBmp(TImage ***&ppp)
{
   if(!ppp) return;
   for (int i=0;i<m_iColLenPre ;i++){
      for (int j=0;j<m_iRowLenPre ;j++)
        delete ppp[i][j];
   }
   for (int i=0;i<m_iColLenPre ;i++){
      delete [] ppp[i];
   }
   delete [] ppp;
   ppp = 0;
}
//-------------------------------------------------------
void _ImgMatrix::fnClearJpg(TJPEGImage ***&ppp)
{
   if(!ppp) return;
   for (int i=0;i<m_iColLenPre ;i++){
      for (int j=0;j<m_iRowLenPre ;j++)
        delete ppp[i][j];
   }
   for (int i=0;i<m_iColLenPre ;i++){
      delete [] ppp[i];
   }
   delete [] ppp;
   ppp = 0;
}
//--------------------------------
void _ImgMatrix::fnSetDisplayPnl(TPanel *pnl)
{
   for (int i=0;i<m_iColLen ;i++){
      for (int j=0;j<m_iRowLen ;j++)
        m_pppBmp[i][j]->Parent= pnl;
   }
}
//------------------------------
void _ImgMatrix::fnSetPos(int iTop, int iLeft, int iW, int iH)
{
   int immW, immH;
   m_iTop = iTop;
   m_iLeft = iLeft;
   if(iW>0) m_iW = iW;
   if(iH>0) m_iH = iH;
   immW = m_iW/m_iColLen;
   immH = m_iH/m_iRowLen;
   for(int i=0; i<m_iColLen; i++){
      for (int j=0; j<m_iRowLen; j++){
         m_pppBmp[i][j]->Width = immW;
         m_pppBmp[i][j]->Height= immH;
         m_pppBmp[i][j]->Top = m_iTop + immH*j;
         m_pppBmp[i][j]->Left= m_iLeft+ immW*i;
      }
   }
}
//---------------------------------------------------
TImage *_ImgMatrix::fnGetBmp(int iCol, int iRow)
{
   if((iCol > (m_iColLen-1)) || (iCol <0)) return 0;
   if((iRow > (m_iRowLen-1)) || (iRow <0)) return 0;
   return m_pppBmp[iCol][iRow];
}
//-------------------------------------------------
TJPEGImage *_ImgMatrix::fnGetJpg(int iCol, int iRow)
{
   if((iCol > (m_iColLen-1)) || (iCol <0)) return 0;
   if((iRow > (m_iRowLen-1)) || (iRow <0)) return 0;
   m_pppJpg[iCol][iRow]->Assign(m_pppBmp[iCol][iRow]->Picture->Bitmap);
   return m_pppJpg[iCol][iRow];
}
//-------------------------------------------------
TMemoryStream *_ImgMatrix::fnGetBmpMem(int iCol, int iRow)
{
   if((iCol > (m_iColLen-1)) || (iCol <0)) return 0;
   if((iRow > (m_iRowLen-1)) || (iRow <0)) return 0;
   m_pppBmpMem[iCol][iRow]->Clear();
  
   m_pppBmp[iCol][iRow]->Picture->Assign(m_pppJpg[iCol][iRow]);

   m_pppBmp[iCol][iRow]->Picture->Bitmap->SaveToStream(m_pppBmpMem[iCol][iRow]);
   return m_pppBmpMem[iCol][iRow];
}
//---------------------------------------------------
TMemoryStream *_ImgMatrix::fnGetJpgMem(int iCol, int iRow)
{
   if((iCol > (m_iColLen-1)) || (iCol <0)) return 0;
   if((iRow > (m_iRowLen-1)) || (iRow <0)) return 0;
   m_pppJpgMem[iCol][iRow]->Clear();

   m_pppJpg[iCol][iRow]->Assign(m_pppBmp[iCol][iRow]->Picture->Bitmap);
   m_pppJpg[iCol][iRow]->CompressionQuality = m_iCmpQual;
   m_pppJpg[iCol][iRow]->SaveToStream(m_pppJpgMem[iCol][iRow]);
   return m_pppJpgMem[iCol][iRow];
}
//------------------------------------------------------
bool _ImgMatrix::fnAssignBmp(TImage *Img, int iCol, int iRow)
{
   if((iCol > (m_iColLen-1)) || (iCol <0)) return false;
   if((iRow > (m_iRowLen-1)) || (iRow <0)) return false;
   m_pppBmp[iCol][iRow]->Picture->Bitmap->Assign(Img->Picture->Graphic);
 //  m_pppBmp[iCol][iRow]->Picture->Assign(Img->Picture);
}
//-------------------------------------------------------
bool _ImgMatrix::fnAssignMtr(_ImgMatrix *Mtr)
{
   TImage *Img;
   bool bRtn=true;
   int iColLen, iRowLen, iW, iH;
   Mtr->fnGetMtrInf(iColLen, iRowLen, iW, iH);
   if((this->m_iColLen != iColLen) || (this->m_iRowLen != iRowLen)) return false;

    for(int i=0; i<m_iColLen; i++){
      for (int j=0; j<m_iRowLen; j++){
         Img = Mtr->fnGetBmp(i,j);
         m_pppBmp[i][j]->Picture->Bitmap->Assign(Img->Picture->Graphic );
         m_pppJpg[i][j]->Assign(m_pppBmp[i][j]->Picture->Bitmap);
      }
   }
   return bRtn;
}
//--------------------------------------------------------
bool _ImgMatrix::fnLoadFromJpgMem(TMemoryStream *mem, int iCol, int iRow)
{
   bool bRtn = true;
   if((iCol > (m_iColLen-1)) || (iCol <0)) return false;
   if((iRow > (m_iRowLen-1)) || (iRow <0)) return false;
   mem->Position=0;
   m_pppJpg[iCol][iRow]->LoadFromStream(mem);
   m_pppBmp[iCol][iRow]->Picture->Assign(m_pppJpg[iCol][iRow]);
   return bRtn;
}
//-------------------------------------------------------
bool _ImgMatrix::fnLoadFromBmpMem(TMemoryStream *mem, int iCol, int iRow)
{
   bool bRtn = true;
   if((iCol > (m_iColLen-1)) || (iCol <0)) return false;
   if((iRow > (m_iRowLen-1)) || (iRow <0)) return false;
   mem->Position=0;
   m_pppBmp[iCol][iRow]->Picture->Bitmap->LoadFromStream(mem);
   m_pppJpg[iCol][iRow]->Assign(m_pppBmp[iCol][iRow]->Picture->Bitmap);
   return bRtn;
}
//-------------------------------------------------------
bool _ImgMatrix::fnSeparateImg(HWND hWndSrc)
{
   bool bRtn =true;
   TImage *ImgTmp;
   int immW, immH;
   int iWidth, iHeight;
   HDC dcS=0;
   RECT rt=TRect(0,0,0,0);
   if(hWndSrc){
      if(IsWindow(hWndSrc)){
         dcS = GetDC(0);// hWnd
         GetWindowRect(hWndSrc, &rt);
         iWidth  = abs(rt.right  - rt.left);
         iHeight = abs(rt.bottom - rt.top);
      }
      else
         return false;
   }
   else{
      dcS = GetDC(0); //hWnd
      iWidth  = m_iSysCx;
      iHeight = m_iSysCy;
   }
   m_iW = iWidth;
   m_iW = iHeight;
   immW = m_iW/m_iColLen;
   immH = m_iH/m_iRowLen;

   for (int i=0;i<m_iColLen ;i++){
      for (int j=0;j<m_iRowLen ;j++){
         ImgTmp = m_pppBmp[i][j];
         ImgTmp->Height=immH;
         ImgTmp->Width =immW;
         BitBlt(ImgTmp->Canvas->Handle ,0 ,0, immW, immH, dcS, rt.left+immW*i, rt.top+immH*j, SRCCOPY);
      }
   }
   ReleaseDC(0, dcS);// hWnd
   return bRtn;
}
//-------------------------------------------------------
void _ImgMatrix::fnSetPlayScl(float fx, float fy)
{
   TImage *ImgTmp;
   for (int i=0;i<m_iColLen ;i++){
      for (int j=0;j<m_iRowLen ;j++){
         ImgTmp = m_pppBmp[i][j];
         ImgTmp->Height=ImgTmp->Height*fx;
         ImgTmp->Width =ImgTmp->Width*fy;
         ImgTmp->Stretch=true;
      }
   }
}
//--------------------------------------------------------
bool _ImgMatrix::fnSeparateImg(TImage *ImgSrc)
{
   bool bRtn = true;
   int immW, immH;
   TImage *ImgTmp;
   if(!ImgSrc) return false;
   if((ImgSrc->Picture->Bitmap->Width==0) || (ImgSrc->Picture->Bitmap->Height==0)) return false;

   m_iW = ImgSrc->Picture->Bitmap->Width;
   m_iH = ImgSrc->Picture->Bitmap->Height;
   immW = m_iW/m_iColLen;
   immH = m_iH/m_iRowLen;
   for (int i=0;i<m_iColLen ;i++){
      for (int j=0;j<m_iRowLen ;j++){
         ImgTmp = m_pppBmp[i][j];
         ImgTmp->Height = immH;
         ImgTmp->Width  = immW;
         BitBlt(ImgTmp->Canvas->Handle ,0 ,0, immW, immH, ImgSrc->Canvas->Handle, immW*i, immH*j, SRCCOPY);
      }
   }
   return bRtn;
}
//--------------------------------------------------------
bool _ImgMatrix::fnCombineImg(TImage *Img, int iCol, int iRow, int iCCLen, int iCRLen)
{
   bool bRtn = true;
   int iDatCol, iDatRow;
   int immW = m_iW/m_iColLen;
   int immH = m_iH/m_iRowLen;
   if(!Img) return false;
   if(iCol > (m_iColLen-1)) iCol = m_iColLen-1;
   if(iCol <0)              iCol = 0;
   if(iRow > (m_iRowLen-1)) iRow = m_iRowLen-1;
   if(iRow <0)              iRow = 0;
   if(iCCLen==-1)           iCCLen = m_iColLen - iCol;
   else{
      if(iCol+iCCLen >= m_iColLen)
         iCCLen = m_iColLen - iCol;
   }
   if(iCRLen==-1)           iCRLen = m_iRowLen - iRow;
   else{
      if(iRow+iCRLen >= m_iRowLen)
         iCRLen = m_iRowLen - iRow;
   }
   Img->Picture->Bitmap->Width  = immW*iCCLen;
   Img->Picture->Bitmap->Height = immH*iCRLen;
   Img->Width  = immW*iCCLen;
   Img->Height = immH*iCRLen;
   TImage *ImgTmp;
   int iW, iH;
   HDC dcD, dcS;
   dcD = Img->Picture->Bitmap->Canvas->Handle;
   for (int i=iCol, M=0; M<iCCLen ;i++, M++){
      for (int j=iRow, N=0; N<iCRLen ;j++, N++){
         ImgTmp = m_pppBmp[i][j];
         iW = ImgTmp->Picture->Bitmap->Width;
         iH = ImgTmp->Picture->Bitmap->Height;
         dcS = ImgTmp->Picture->Bitmap->Canvas->Handle;
         BitBlt(dcD ,iW*M, iH*N, iW, iH, dcS, 0, 0, SRCCOPY);
      }
   }
   Img->Refresh();
   return bRtn;
}
//---------------------------------------------------------
_Tank::_Tank(int iVol)
{
   m_iVol = iVol;
   m_szBuff = new char[m_iVol];
   fnClear();
}
//--------------------------
_Tank::~_Tank()
{
   delete [] m_szBuff;
}
//-----------------------------------------------------------------------
void _Tank::fnClear()
{
   ::ZeroMemory(m_szBuff, sizeof(char) * m_iVol);
   m_iBuffPos = 0;
   m_iTankMagn = m_iVol/10;
}
//-------------------------------------------------------------
int _Tank::fnPush(void *pData, int &iLen)
{
   int iRtn;
   memcpy((void *)(m_szBuff+m_iBuffPos),pData, iLen);
   m_iBuffPos += iLen;
   if((m_iBuffPos+m_iTankMagn) >= m_iVol){
      iRtn = -1; //ֺF
   }
   else{
      iRtn = 0; //ٮtܻ
   }

   return iRtn;
}
//---------------------------------------------------------
void _Tank::fnDump(void *&pData, int &iLen)
{
   pData = (void *)(m_szBuff);
   iLen = m_iBuffPos;
   m_iBuffPos = 0;
}
//-----------------------------------------------------------
int _Tank::fnGetLen()
{
   return  m_iBuffPos;
}
//---------------------------------------------------------------
int  _Tank::fnPull(void *&pData)
{
   pData = (void *)(m_szBuff);
   return m_iBuffPos;
}
//---------------------------------------------------------------
int _Tank::fnPull(void *&pData,int iLen)
{
   int iPullLen=0;
   if(iLen >= m_iVol)
      iPullLen = m_iVol;
   pData = (void *)(m_szBuff);
   return  iPullLen;
}
//--------------------------------------------------------
